
"""
zt_errors.py
--------------

The module contains the errors that can be raised by the zttt package.
"""


class ZTError(Exception):
    """Base class for all zt_errors."""
    pass


class ZTBadFunctionCall(AttributeError, ZTError):
    """Raised when a function is called with the wrong arguments. Mostly used for debugging."""
    pass


class ZTGameException(ZTError):
    """Raised when a function call crashes the game."""
    pass


class ZTWrongInput(ValueError, ZTGameException):
    """Raised when the input provided is not valid."""
    pass



from typing import Dict, Tuple, List

from ..zt_base import ZTBaseEngine
from random import choice


class ZTPlayerFirst (ZTBaseEngine):
    """Class for the Engine

    The class is responsible for the logic when player plays first
    """

    # Engine Move for Move 2
    __ENGINE_MOVE: Dict[Tuple[int, int], List[int]] = {
        (0, 8): [1, 3, 5, 7], (2, 6): [1, 3, 5, 7],  # Opposite corner cases
        (1, 7): [0, 2, 6, 8], (3, 5): [0, 2, 6, 8],  # Opposite edge cases
        (1, 3): [0], (1, 5): [2], (7, 3): [6], (7, 5): [8],  # Non-opposite edge cases
    }

    def __init__(self) -> None:
        """Initialize the State Variables"""
        ZTBaseEngine.__init__(self, False)
        self.center: bool = False  # This variable is True when the opponent plays center in move 1
        self.liberty_move3: bool = False  # This means choice of the third move is corners

    def __get_non_center_move2(self) -> int:
        """Returns the second move using the __ENGINE_MOVE dictionary

        :return: The move for move 2
        :rtype: int
        """

        for (pos1, pos2) in ZTPlayerFirst.__ENGINE_MOVE:
            if self._board_list[pos1] == self._player_value and self._board_list[pos2] == self._player_value:
                return choice(ZTPlayerFirst.__ENGINE_MOVE[(pos1, pos2)])

        non_center_lines: List[Tuple[int, int, int]] = [(0, 1, 2), (2, 5, 8), (6, 7, 8), (0, 3, 6)]
        for line in non_center_lines:
            line_weight = self._calc_line_weight(line)
            if line_weight == 2 * self._player_value:
                # One of the corner is empty implies a continuous draw game to the last move
                if self._board_list[line[0]] == self._VAL_EMPTY:
                    return line[0]
                if self._board_list[line[2]] == self._VAL_EMPTY:
                    return line[2]

                # The edge is empty introduces a liberty move in move 3
                self.liberty_move3 = True
                return line[1]

        # The remaining possibility is that the player has played in one corner and one edge
        # We just play opposite of the corner the player has played in

        # The remaining cases
        if self._board_list[0] == self._player_value:
            return 8
        if self._board_list[2] == self._player_value:
            return 6
        if self._board_list[6] == self._player_value:
            return 2
        if self._board_list[8] == self._player_value:
            return 0

    def play(self, inp: int) -> None:
        """Plays the player's move at the position specified

        :param inp: The position to play the player's move
        :type inp: int
        :return: None
        """

        self._play_player(inp)

        if not self.status:
            return

        if self.move == 2:
            if inp == 4:
                self.center = True  # The player played the center so the best move is a corner
                return self._play_engine(choice(self.empty_corners))
            return self._play_engine(4)  # If the player does not start with center, the engine will

        if self.move == 4 and not self.center:
            return self._play_engine(self.__get_non_center_move2())

        if self.move == 4 and self.center:
            move = self._get_danger_move()
            if move:
                return self._play_engine(move[0])
            return self._play_engine(choice(self.empty_corners))

        # Special Cases Over
        move = self._get_bot_move()
        if move:
            return self._play_engine(move[0])

        if self.move == 6:
            if self.liberty_move3:
                return self._play_engine(choice(self.empty_edges))
            else:
                return self._play_engine(choice(self.empty_corners))

        if self.move == 8:
            return self._play_engine(choice(self._empty_positions))



from .zt_base import ZTBaseBoard


class PvP(ZTBaseBoard):

    def __init__(self):
        super().__init__()

    def play(self, pos: int):
        if self.turn == 1:
            super()._play_player_one_move(pos)
        else:
            super()._play_player_two_move(pos)



from typing import Dict, List
from random import choice

from ..zt_base import ZTBaseEngine


class ZTEngineFirst(ZTBaseEngine):
    """Class for the Engine

    The class is responsible for the logic when engine plays first
    """

    __ENGINE_MOVE: Dict[int, Dict[int, List[int]]] = {
        0: {1: [6, 4], 2: [6, 8], 3: [2, 4], 4: [5, 7], 5: [2, 4, 6], 6: [2, 8], 7: [2, 4, 6], 8: [2, 6]},
        2: {0: [6, 8], 1: [8, 4], 3: [0, 4, 8], 4: [3, 7], 5: [0, 4], 6: [0, 8], 7: [0, 4, 8], 8: [0, 6]},
        6: {0: [2, 8], 1: [0, 4, 8], 2: [0, 8], 3: [8, 4], 4: [1, 5], 5: [0, 4, 8], 7: [0, 4], 8: [0, 2]},
        8: {0: [2, 6], 1: [2, 4, 6], 2: [0, 6], 3: [2, 4, 6], 4: [1, 3], 5: [6, 4], 6: [0, 2], 7: [2, 4]}
    }

    def __init__(self):
        """Initialize the State Variables"""
        ZTBaseEngine.__init__(self, True)
        self.move1: int = choice([0, 2, 6, 8])
        self._play_engine(self.move1)

    # A function which is called in the main method if play is True
    def play(self, pos: int) -> None:
        """Plays the player's move at the position specified

        :param pos: The position to play the player's move
        :type pos: int
        :return: None
        """

        self._play_player(pos)

        if self.move == 3:
            return self._play_engine(choice(ZTEngineFirst.__ENGINE_MOVE[self.move1][pos]))

        move = self._get_bot_move()
        if move:
            return self._play_engine(move[0])

        if self.empty_corners:
            return self._play_engine(choice(self.empty_corners))

        else:
            return self._play_engine(choice(self._empty_positions))



from .zt_engines import ZTEngineFirst
from .zt_engines import ZTPlayerFirst


class PvC(ZTEngineFirst, ZTPlayerFirst):
    """Class for the PvC Game"""

    def __init__(self, _engine_first: bool = True) -> None:
        """Initialize the Game

        :param _engine_first: Specifies if the engine starts first
        :type _engine_first: bool
        :return: None
        """

        if _engine_first:
            self.parent = ZTEngineFirst
            ZTEngineFirst.__init__(self)

        else:
            self.parent = ZTPlayerFirst
            ZTPlayerFirst.__init__(self)

    def play(self, pos: int) -> None:
        self.parent.play(self, pos)



from typing import Tuple, List
from random import choice

from .zt_base_board import ZTBaseBoard
from ..zt_errors import *


class ZTBaseEngine(ZTBaseBoard):
    """Base Class for the Engine"""

    def __init__(self, _engine_first: bool) -> None:
        """Initialize the State Variables in an Engine

        :param _engine_first: Specifies if the engine starts first
        :type _engine_first: bool
        :return: None
        """

        super().__init__()

        # Private since we don't want people to change it
        self.__engine_first: bool = _engine_first

        # List of all the lines that contain at least one empty position
        self.__active_lines: List[Tuple[int, int, int]] = [
            (0, 1, 2), (3, 4, 5), (6, 7, 8),
            (0, 3, 6), (1, 4, 7), (2, 5, 8),
            (0, 4, 8), (2, 4, 6)
        ]

    # Properties

    @property
    def engine_first(self) -> bool:
        """Returns if the engine starts first

        :return: True if the engine starts first, False otherwise
        :rtype: bool
        """
        return self.__engine_first

    @property
    def _engine_value(self) -> int:
        """Returns the value of the engine and is mainly for development purpose

        :return: The value of the engine
        :rtype: int
        """
        return self.__class__._VAL_PLAYER1 if self.__engine_first is True else self.__class__._VAL_PLAYER2

    @property
    def _player_value(self) -> int:
        """Returns the value of the player and is mainly for development purpose

        :return: The value of the engine
        :rtype: int
        """
        return self.__class__._VAL_PLAYER1 if self.__engine_first is False else self.__class__._VAL_PLAYER2

    # Bot Functions

    def _calc_line_weight(self, _line: Tuple[int, int, int], _board_list: List[int] = None) -> int:
        """Calculates the weight of a line

        :param _line: The line to calculate the weight of
        :type _line: Tuple[int, int, int]
        :param _board_list: If specified, the board list to use instead of the current one
        :type _board_list: List[int]
        :return: The weight of the line (Sum of the values of the positions)
        :rtype: int
        :raises ZTBadFunctionCall: If the board list is not in proper format
        """

        if _board_list is None:
            _board_list = self._board_list
        elif len(_board_list) != 9:
            raise ZTBadFunctionCall("The Board List length must be 9")
        else:
            for pos_val in _board_list:
                if pos_val not in (self._VAL_PLAYER1, self._VAL_PLAYER2, self._VAL_EMPTY):
                    raise ZTBadFunctionCall(f"Board List not in the correct format, {pos_val} invalid")

        # The verification is complete
        return sum([_board_list[pos] for pos in _line])

    def _get_winnable_moves(self) -> List[int]:
        """Returns a list of winnable moves and empty list if none

        :return: A list of winnable moves
        :rtype: List[int]
        """

        winnable_moves = []
        for line in self.__active_lines:
            line_weight = self._calc_line_weight(line)
            if line_weight == 2 * self._engine_value:
                if self._board_list[line[0]] == self._VAL_EMPTY:
                    winnable_moves.append(line[0])
                elif self._board_list[line[1]] == self._VAL_EMPTY:
                    winnable_moves.append(line[1])
                else:
                    winnable_moves.append(line[2])

        return winnable_moves

    def _get_danger_move(self) -> List[int]:
        """Returns the move to avoid losing, else nothing

        :return: The singleton list with the only move to avoid losing, empty list otherwise
        :rtype: List[int]
        """

        for line in self.__active_lines:
            line_weight = self._calc_line_weight(line)
            if line_weight == 2 * self._player_value:
                if self._board_list[line[0]] == self._VAL_EMPTY:
                    return [line[0]]
                elif self._board_list[line[1]] == self._VAL_EMPTY:
                    return [line[1]]
                else:
                    return [line[2]]
        return []

    def __get_double_danger_moves(self) -> List[int]:
        """Returns a list of moves that cause double attacks, empty list if none

        :return: List of moves that cause double danger
        :rtype: List[int]
        """

        double_danger_moves = []
        for corner in self.empty_corners:
            count = 0
            new_board_list = self.board_list
            new_board_list[corner] = self._engine_value
            for line in self.__active_lines:
                line_weight = self._calc_line_weight(line, new_board_list)
                if line_weight == 2 * self._engine_value:
                    count += 1
                if count == 2:
                    break
            if count == 2:
                double_danger_moves.append(corner)
        return double_danger_moves

    def _get_bot_move(self) -> List[int]:
        """Returns the best move to play using the above functions

        :return: The singleton list with the only move to play, empty list otherwise
        :rtype: List[int]
        """

        temp = self._get_winnable_moves()
        if temp:
            return [choice(temp)]

        temp = self._get_danger_move()
        if temp:
            return [choice(temp)]

        temp = self.__get_double_danger_moves()
        if temp:
            return [choice(temp)]

        return []

    def __clean(self, pos: int) -> None:
        """Clean the active lines list maintained by the engine

        :param pos: The last position played
        :return: None
        """

        if self.move < 5:
            return
        for (pos1, pos2) in self.LINES[pos]:
            if sorted([pos, pos1, pos2]) in self.__active_lines and self._board_list[pos1] != self._VAL_EMPTY \
                    and self._board_list[pos2] != self._VAL_EMPTY:
                self.__active_lines.remove(sorted([pos, pos1, pos2]))

    # Functions to play

    def _play_engine(self, pos: int) -> None:
        """Play the engine move. Also cleans the active lines list

        :param pos: Position for engine to play
        :return: None
        """

        self.__clean(pos)
        if self.__engine_first:
            self._play_player_one_move(pos)
        else:
            self._play_player_two_move(pos)

    def _play_player(self, pos: int) -> None:
        """Play the player move. Also cleans the active lines list

        :param pos: Position for player to play
        :return: None
        """

        self.__clean(pos)
        if not self.__engine_first:
            self._play_player_one_move(pos)
        else:
            self._play_player_two_move(pos)



from typing import List, Tuple, Dict, Callable
from ..zt_errors import *


class ZTBaseBoard:
    """This is the base class which is inherited by all the classes which are used in the module"""

    # The following variables are not to be changed
    _VAL_PLAYER1 = 4  # Value for player 1 in the board_list
    _VAL_PLAYER2 = 1  # Value for player 2 in the board_list
    _VAL_EMPTY = 0  # Value for empty position in the board_list

    # Indicates what the values contained in the _board_list represent
    # The indicator with value 5 goes first
    _INDICATOR: Dict[int, str] = {
        _VAL_PLAYER1: 'X',
        _VAL_PLAYER2: 'O',
        _VAL_EMPTY: ' '
    }

    # All the lines the current position
    LINES: Dict[int, List[Tuple[int, int]]] = {
        4: [(0, 8), (2, 6), (1, 7), (3, 5)],

        0: [(1, 2), (4, 8), (3, 6)],
        2: [(1, 0), (4, 6), (5, 8)],
        6: [(3, 0), (4, 2), (7, 8)],
        8: [(5, 2), (4, 0), (6, 7)],

        1: [(0, 2), (4, 7)],
        3: [(0, 6), (4, 5)],
        5: [(2, 8), (3, 4)],
        7: [(6, 8), (4, 1)]
    }

    # The center is present in Position 4
    _CENTER = 4

    @classmethod
    def set_indicators(cls, _player1: str = 'X', _player2: str = 'O', _space: str = ' ') -> None:
        """Sets the indicators

        :param _player1: Indicator for player 1
        :type _player1: str
        :param _player2: Indicator for player 2
        :type _player2: str
        :param _space: Indicator for empty position
        :type _space: str
        :return: None
        """

        cls._INDICATOR = {
            cls._VAL_PLAYER1: str(_player1),
            cls._VAL_PLAYER2: str(_player2),
            cls._VAL_EMPTY: str(_space)
        }

    def __init__(self) -> None:
        """Initializes the board and the state variables"""

        # Main list containing the _empty_positions from top left in row major order
        # Protected since inherited classes will be able to quickly access it
        self._board_list: List[int] = [0 for _ in range(9)]

        # List containing the _empty_positions empty edges and empty corners
        # Protected since inherited classes will be able to quickly access it
        self._empty_positions: List[int] = [i for i in range(9)]

        # History
        self._history: List[int] = []

        # Determines whether the game is ongoing or not
        self.__status: bool = True

        # Stores the information about the __winner of the game
        self.__winner: (None, int) = None  # Initially set to None since there is no winner

        # This stores the current move that is to be performed
        # move = 1 indicates that move 1 is to be played now
        self.__move: int = 1

        # Stores the values that are to be highlighted after the game
        self.__highlight_list: List[int] = list()

        # Event triggers
        self._on_move: Callable[[int, int], None] = lambda player, pos: None
        self._on_finish: Callable[[int], None] = lambda player: None

    def __repr__(self) -> str:
        """Returns the string representation of the board"""
        return self.board

    @property
    def board_list(self) -> List[int]:
        """List containing the status of the board (Row Major Order)

        :return: A (duplicate) list of the board list
        :rtype: List[int]
        """
        return self._board_list[:]

    @property
    def status(self) -> bool:
        """Whether the game is in progress or not

        :return: The status of the game
        :rtype: bool
        """
        return self.__status

    @property
    def move(self) -> int:
        """The current move that is to be performed

        :return: Current Move Number
        :rtype: int
        """
        return self.__move

    @property
    def turn(self) -> int:
        """The player whose turn it is to play

        :return: Player Number
        :rtype: int
        """
        return 1 if self.__move % 2 == 1 else 2

    @property
    def winner(self) -> (None, int):
        """The winner of the game

        :return: Winner if the game is over, 0 is draw, None if no winner yet
        :rtype: (None, int)
        """
        # !TODO Issue a Warning if there is no winner
        if self.status:
            pass  # Issue a warning here
        return self.__winner

    @property
    def empty_positions(self) -> List[int]:
        """List containing the empty positions of the board

        :return: A (duplicate) list (in row major order from 0) of the empty positions
        :rtype: List[int]
        """
        return self._empty_positions[:]

    @property
    def empty_corners(self) -> List[int]:
        """List containing the empty corners of the board

        :return: List of the empty corners
        :rtype: List[int]
        """
        return [i for i in self._empty_positions if i is not self._CENTER and i % 2 == 0]

    @property
    def empty_edges(self) -> List[int]:
        """List containing the empty edges of the board

        :return: List of the empty edges
        :rtype: List[int]
        """
        return [i for i in self._empty_positions if i % 2 == 1]

    @property
    def highlighted(self) -> List[int]:
        """Returns the highlighted positions of the board when there is a winner

        :return: A (duplicate) list of highlighted positions
        :rtype: List[int]
        """
        return self.__highlight_list[:]

    # Some functions for printing the board

    @property
    def board(self) -> str:
        """Returns the string representation of the board

        :return: board string
        """

        pr_str = ' _____ _____ _____\n'
        for i in range(3):
            pr_str += '|     |     |     |\n|'
            for j in range(3):
                pos_val = self._board_list[i * 3 + j]
                pr_str += f'  {self.__class__._INDICATOR[pos_val]}  |'
            pr_str += '\n|_____|_____|_____|\n'
        return pr_str

    @property
    def on_move(self):
        """Returns the on_move event trigger

        :return: on_move event trigger
        :rtype: Callable[[int], None]
        """
        return self._on_move

    @on_move.setter
    def on_move(self, on_move: Callable[[int, int], None]):
        """Sets the on_move event trigger

        :param on_move: The on_move callable
        :type on_move: Callable[[int, int], None]
        :raise: TypeError if on_move is not a callable
        """

        if type(on_move) is not callable:
            raise TypeError('on_move must be a function')
        self._on_move = on_move

    @property
    def on_finish(self):
        """Returns the on_move event trigger

        :return: on_finish event trigger
        :rtype: Callable[[int], None]
        """
        return self._on_finish

    @on_finish.setter
    def on_finish(self, on_finish: Callable[[int], None]):
        """Sets the on_move event trigger

        :param on_finish: The on_finish callable
        :type on_finish: Callable[[int], None]
        :raise: TypeError if on_finish is not a callable
        """
        if type(on_finish) is not callable:
            raise TypeError('on_move must be a function')
        self._on_finish = on_finish

    # Helper Functions

    def _verify_status(self) -> None:
        """Raises an error if the status is not true

        :raise: ZTGameException if the status is False
        """
        if not self.__status:
            raise ZTGameException("Game is not in progress")

    def _verify_pos(self, pos: int) -> None:
        """Raises an error if pos is not empty

        :param pos: The position to be verified
        :type pos: int
        :raise: ZTGameException if the position is not empty
        """
        if not self._board_list[pos] == 0:
            raise ZTGameException("Position Already Taken")

    # Returns True if a tuple of board _empty_positions does not contain an empty position
    def __is_fully_played(self, _line: Tuple[int, int, int]) -> bool:
        """Checks if the given line is fully played

        :param _line: A tuple of three positions
        :type _line: Tuple[int, int, int]
        :return: True if the tuple of positions is fully played, False otherwise
        :rtype: bool
        """
        for pos in _line:
            if self._board_list[pos] == 0:
                return False
        return True

    def __check_win(self, pos: int) -> bool:
        """Checks if a player won after the current move

        :param pos: The position that was just played
        :type pos: int
        :return: True if the player won, False otherwise
        :rtype: bool
        """

        if self.__move < 5:
            return False

        winning_set_list = [pos]
        for line in ZTBaseBoard.LINES[pos]:
            line_val = self._board_list[pos] + \
                       self._board_list[line[0]] + self._board_list[line[1]]

            if line_val in (3 * ZTBaseBoard._VAL_PLAYER1, 3 * ZTBaseBoard._VAL_PLAYER2):
                winning_set_list.extend(line)

        if len(winning_set_list) > 1:
            self.__highlight_list = winning_set_list
            return True

        return False

    # Player Moves

    def __play_player_move(self, player: int, pos: int) -> bool:
        """Plays a move for the player at the given position

        :param player: The player who is making the move
        :type player: int
        :param pos: The position to be played
        :type pos: int
        :return: True if there was a winner, False otherwise
        :rtype: bool
        :raise: ZTGameException if the game is not in progress
        :raise: ZTGameException if the move is invalid
        """

        self._verify_status()
        self._verify_pos(pos)

        if self.__move % 2 != player % 2:
            raise ZTGameException("It is currently not the player's move")

        self._board_list[pos] = ZTBaseBoard._VAL_PLAYER1 if player == 1 else ZTBaseBoard._VAL_PLAYER2
        self.on_move(player, pos)
        self._empty_positions.remove(pos)
        self._history.append(pos)

        win = self.__check_win(pos)
        self.__move += 1  # Increasing the move by 1
        return win

    def _play_player_one_move(self, pos: int) -> None:
        """Function for handling the player 1 move

        :param pos: The position to be played
        :type pos: int
        :return: None
        :raise: ZTGameException if the game is not in progress
        :raise: ZTGameException if the move is invalid
        """
        win = self.__play_player_move(1, pos)

        if win:
            self.__finisher(1)  # Player 1 wins
        elif self.__move == 10:
            self.__finisher(0)  # No one Wins and 0 is a draw

    def _play_player_two_move(self, pos: int) -> None:
        """Function for handling the player 2 move

        :param pos: The position to be played
        :type pos: int
        :return: None
        :raise: ZTGameException if the game is not in progress
        :raise: ZTGameException if the move is invalid
        """
        win = self.__play_player_move(2, pos)

        if win:
            self.__finisher(2)

    def __finisher(self, winner: int) -> None:
        """Function for handling the end of the game

        :param winner: The winner of the game
        :type winner: int
        :return: None
        """
        self.__winner = winner
        self.__status = False
        self.on_finish(winner)
